/**
 ******************************************************************************
 * @copyright Copyright (c) 2020, StarFive Technology Co.,Ltd. All rights reserved.
 * 
 * @file crc32.c
 * @author StarFive FW Team
 * @brief 
 ******************************************************************************
 */
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

static uint32_t g_crc, g_busy;

uint32_t crc32_mpeg(const char *buf, int len)
{
    uint32_t i;
    uint32_t crc = 0xFFFFFFFF;
    
    while(len--)
    {
        crc ^= (uint32_t)(*buf++) << 24;
        for (i = 0; i < 8; ++i)
        {
            if (crc & 0x80000000)
                crc = (crc << 1) ^ 0x04C11DB7;
            else
                crc <<= 1;
        }
    }
    return crc;
}

uint32_t crc32(uint32_t crc, const char *buf, size_t len)
{
	static uint32_t table[256];
	static int have_table = 0;
	uint32_t rem;
	uint8_t octet;
	int i, j;
	const char *p, *q;

	/* This check is not thread safe; there is no mutex. */
	if (have_table == 0) {
		/* Calculate CRC table. */
		for (i = 0; i < 256; i++) {
			rem = i;  /* remainder from polynomial division */
			for (j = 0; j < 8; j++) {
				if (rem & 1) {
					rem >>= 1;
					rem ^= 0xedb88320;
				} else
					rem >>= 1;
			}
			table[i] = rem;
		}
		have_table = 1;
	}
 
	crc = ~crc;
	q = buf + len;
	for (p = buf; p < q; p++) {
		octet = *p;  /* Cast to unsigned octet. */
		crc = (crc >> 8) ^ table[(crc & 0xff) ^ octet];
	}
	return ~crc;
}

int crc32_init(void)
{
    if (g_busy) {
        return -EBUSY;
    }
    g_crc = 0;
    return 0;
}

int crc32_update(const char *buf, size_t len)
{
    g_crc = crc32(g_crc, buf, len);
    return 0;
}

int crc32_update_file(const char *path)
{
    FILE *fp = fopen(path, "rb");
    if (!fp) {
        return -EIO;
    }

    const uint32_t BUF_SIZE = 1024;
    char buf[BUF_SIZE];
    size_t rd_len = 0; 
    while ((rd_len = fread(buf, 1, BUF_SIZE, fp)) > 0) {
        crc32_update(buf, rd_len);
    }
    fclose(fp);
    return 0;
}

int crc32_final(uint32_t *crc)
{
    if (crc) {
        *crc = g_crc;
    }
    g_busy = 0;
    return 0;
}

int crc32_file(const char *path, uint32_t *crc)
{
    if (crc32_init()) {
        return -EBUSY;
    }
    FILE *fp = fopen(path, "rb");
    if (!fp) {
        return -EIO;
    }

    const uint32_t BUF_SIZE = 1024;
    char buf[BUF_SIZE];
    size_t rd_len = 0; 
    while ((rd_len = fread(buf, 1, BUF_SIZE, fp)) > 0) {
        crc32_update(buf, rd_len);
    }
    fclose(fp);
    crc32_final(crc);
    return 0;
}
